import inspect
import logging
import os

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger('jaxns')


def get_grandparent_info(relative_depth: int = 7):
    """
    Get the file, line number and function name of the caller of the caller of this function.

    Args:
        relative_depth: the number of frames to go back from the caller of this function. Default is 6. Should be
        enough to get out of a jax.tree.map call.

    Returns:
        str: a string with the file, line number and function name of the caller of the caller of this function.
    """
    # Get the grandparent frame (caller of the caller)
    s = []
    for depth in range(1, min(1 + relative_depth, len(inspect.stack()) - 1) + 1):
        caller_frame = inspect.stack()[depth]
        caller_file = caller_frame.filename
        caller_line = caller_frame.lineno
        caller_func = caller_frame.function
        s.append(f"{os.path.basename(caller_file)}:{caller_line} in {caller_func}")
    s = s[::-1]
    s = f"at {' -> '.join(s)}"
    return s



import jax
import numpy as np
import pytest
from jax import numpy as jnp

from jaxns.framework.context import wrap_random, get_parameter, get_state, global_context, transform_with_state, \
    next_rng_key, scope, ScopedDict


def test_context():
    def my_model(x):
        # Example of parameter usage
        w_init = wrap_random(jax.nn.initializers.glorot_normal())
        w = get_parameter("w", (x.shape[-1], 128), init=w_init)

        # Example of state usage
        reg_loss = 0.01 * jnp.sum(w ** 2)
        get_state("reg_loss", init=reg_loss)

        param_pytree = get_parameter("param_pytree", init={"a": jnp.ones(()), "b": jnp.ones(())})
        state_pytree = get_state("state_pytree", init=lambda *args: {"a": jnp.ones(()), "b": jnp.ones(())})

        return jnp.dot(x, w) + param_pytree["a"] + param_pytree["b"] + reg_loss + state_pytree["a"] + state_pytree["b"]

    assert len(global_context.stack) == 0

    # Transform the model
    init, apply = transform_with_state(my_model)
    assert len(global_context.stack) == 0

    # Initialize parameters and state
    rng = jax.random.PRNGKey(42)
    x = jnp.ones([1, 32])
    f0, params, states = init(rng, x)

    # Apply the model
    f1, states1 = apply(params, states, jax.random.PRNGKey(0), x)

    np.testing.assert_allclose(f1, f0)
    np.testing.assert_allclose(states['reg_loss'], states1['reg_loss'])

    with pytest.raises(ValueError, match="No context available."):
        _ = get_parameter("w", init=jnp.ones((1, 2)))
    with pytest.raises(ValueError, match="No context available."):
        _ = get_state("reg_loss", init=jnp.ones((1, 2)))
    with pytest.raises(ValueError, match="No context available."):
        next_rng_key()


def test_scope():
    def f():
        x = get_parameter("x", init=jnp.ones(()))
        x_state = get_state("X", init=jnp.ones(()))
        with scope('a'):
            y = get_parameter("y", init=jnp.ones(()))
            with scope('b'):
                z = get_parameter("z", init=jnp.ones(()))
                return x + y + z

    transformed_fn = transform_with_state(f)
    init = transformed_fn.init(jax.random.PRNGKey(42))
    print(init)
    assert '.x' in init.params.keys()
    assert '.X' in init.states.keys()
    assert 'a.y' in init.params.keys()
    assert 'a.b.z' in init.params.keys()
    output = jax.jit(transformed_fn.apply)(init.params, init.states, jax.random.PRNGKey(42))
    print(output)


def test_scope_dict_flatten():
    sd = ScopedDict()
    sd.push_scope("scope")
    sd["key"] = 1
    leaves, treedef = jax.tree_util.tree_flatten(sd)
    sd2 = jax.tree_util.tree_unflatten(treedef, leaves)
    assert sd2["key"] == 1
    assert sd2.scopes == ["scope"]

    sd = ScopedDict()
    sd.push_scope("scope")
    sd["key"] = jnp.ones(1)

    def f(sd):
        return sd


    f_jit = jax.jit(f).lower(sd).compile()
    sd2 = f_jit(sd)
    np.testing.assert_allclose(sd2["key"], jnp.ones(1))
    assert sd2.scopes == ["scope"]



from typing import NamedTuple, Any, Tuple

import jax
from jax import numpy as jnp, random, lax

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import PRNGKey, UType, MeasureType, IntArray
from jaxns.nested_samplers.common.types import Sample


def _single_uniform_sample(key: PRNGKey, model: BaseAbstractModel) -> Sample:
    """
    Gets a single sample strictly within -inf bound (the entire prior), accounting for forbidden regions.

    Args:
        key: PRNGKey
        model: the model to use.

    Returns:
        a sample
    """

    log_L_constraint = jnp.asarray(-jnp.inf, mp_policy.measure_dtype)

    class CarryState(NamedTuple):
        key: PRNGKey
        U: UType
        log_L: MeasureType
        num_likelihood_evals: IntArray

    def cond(carry_state: CarryState):
        return carry_state.log_L <= log_L_constraint

    def body(carry_state: CarryState) -> CarryState:
        key, sample_key = random.split(carry_state.key, 2)
        U = model.sample_U(key=sample_key)
        log_L = model.forward(U=U)
        num_likelihood_evals = carry_state.num_likelihood_evals + jnp.ones_like(carry_state.num_likelihood_evals)
        return CarryState(key=key, U=U, log_L=log_L, num_likelihood_evals=num_likelihood_evals)

    key, sample_key = random.split(key, 2)
    init_U = model.sample_U(key=sample_key)
    init_log_L = model.forward(init_U)
    init_carry_state = CarryState(
        key=key,
        U=init_U,
        log_L=init_log_L,
        num_likelihood_evals=jnp.asarray(1, mp_policy.count_dtype)
    )

    carry_state = lax.while_loop(cond_fun=cond, body_fun=body, init_val=init_carry_state)

    sample = Sample(
        U_sample=carry_state.U,
        log_L_constraint=log_L_constraint,
        log_L=carry_state.log_L,
        num_likelihood_evaluations=carry_state.num_likelihood_evals
    )
    return sample


def draw_uniform_samples(keys: PRNGKey, model: BaseAbstractModel, method: str = 'vmap') -> Sample:
    """
    Get initial live points from uniformly sampling the entire prior.

    Args:
        keys: PRNGKey
        model: the model
        method: which way to draw the init points. vmap is vectorised, and for performant but uses more memory.

    Returns:
        uniformly drawn samples within -inf bound
    """

    if method == 'vmap':
        return jax.vmap(lambda _key: _single_uniform_sample(key=_key, model=model))(keys)
    elif method == 'scan':

        def body(carry_unused: Any, key: PRNGKey) -> Tuple[Any, Sample]:
            return carry_unused, _single_uniform_sample(key=key, model=model)

        _, samples = lax.scan(body, (), keys)

        return samples
    else:
        raise ValueError(f'Invalid method {method}')



from abc import ABC, abstractmethod
from typing import TypeVar, Tuple, NamedTuple, Generic, Any

from jaxns.internals.types import PRNGKey, FloatArray
from jaxns.nested_samplers.common.types import Sample, TerminationRegister, LivePointCollection

SamplerState = TypeVar('SamplerState')


class EphemeralState(NamedTuple):
    # Contains all available ephemeral quantities that might be used for constructing a sampler state.
    # Excludes the full state, as this is too large oft times.
    key: PRNGKey
    live_points_collection: LivePointCollection
    termination_register: TerminationRegister


class AbstractSampler(ABC, Generic[SamplerState]):

    @abstractmethod
    def _pre_process(self, ephemeral_state: EphemeralState) -> Any:
        """
        Run this periodically on the state to produce a data pytree that can be used by the sampler, and
        updated quickly.

        Args:
            ephemeral_state: the current state of the sampler

        Returns:
            any valid pytree
        """
        ...

    def pre_process(self, ephemeral_state: EphemeralState) -> SamplerState:
        """
        Run this periodically on the state to produce a data pytree that can be used by the sampler, and
        updated quickly.

        Args:
            ephemeral_state: the current state of the sampler

        Returns:
            any valid pytree
        """
        return self._pre_process(ephemeral_state)

    @abstractmethod
    def _post_process(self, ephemeral_state: EphemeralState,
                      sampler_state: Any) -> Any:
        """
        Post process the sampler state, after the sampler has been run. Should be quick.

        Args:
            ephemeral_state: a sample collection post sample step
            sampler_state: data pytree produced by the sampler

        Returns:
            the updated sampler state
        """
        ...

    def post_process(self, ephemeral_state: EphemeralState,
                     sampler_state: SamplerState) -> SamplerState:
        """
        Post process the sampler state, after the sampler has been run. Should be quick.

        Args:
            ephemeral_state: a sample collection post sample step
            sampler_state: data pytree produced by the sampler

        Returns:
            the updated sampler state
        """
        return self._post_process(ephemeral_state, sampler_state)

    @abstractmethod
    def _get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: Any) -> Tuple[
        Sample, Sample]:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            sampler_state: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample, and phantom samples
        """
        ...

    def get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: SamplerState) -> Tuple[
        Sample, Sample]:
        """
        Produce a single i.i.d. sample from the model within the log_L_constraint.

        Args:
            key: PRNGkey
            log_L_constraint: the constraint to sample within
            sampler_state: the data pytree needed and produced by the sampler

        Returns:
            an i.i.d. sample, and phantom samples
        """
        return self._get_sample(key, log_L_constraint, sampler_state)

    @abstractmethod
    def num_phantom(self) -> int:
        """
        The number of phantom samples produced by the sampler.

        Returns:
            number of phantom samples
        """
        ...



import jax
import jax.numpy as jnp


def tree_dot(x, y):
    dots = jax.tree.leaves(jax.tree.map(jnp.vdot, x, y))
    return sum(dots[1:], start=dots[0])


def tree_norm(x):
    norm2 = tree_dot(x, x)
    if jnp.issubdtype(norm2.dtype, jnp.complexfloating):
        return jnp.sqrt(norm2.real)
    return jnp.sqrt(norm2)


def tree_mul(x, y):
    return jax.tree.map(jax.lax.mul, x, y)


def tree_sub(x, y):
    return jax.tree.map(jax.lax.sub, x, y)


def tree_div(x, y):
    return jax.tree.map(jax.lax.div, x, y)



from functools import partial
from typing import Optional, Tuple

import jax
from jax import numpy as jnp, random
from jax._src.mesh import Mesh
from jax._src.partition_spec import PartitionSpec
from jax.experimental.shard_map import shard_map

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.maps import tree_device_put
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.shrinkage_statistics import create_init_evidence_calc
from jaxns.internals.types import PRNGKey
from jaxns.nested_samplers.common.types import StaticStandardSampleCollection, \
    NestedSamplerState, TerminationRegister, LivePointCollection
from jaxns.nested_samplers.common.uniform_sample import draw_uniform_samples


def create_init_state(key: PRNGKey, num_live_points: int, max_samples: int, model: BaseAbstractModel,
                      mesh: Optional[Mesh] = None) -> Tuple[LivePointCollection, NestedSamplerState]:
    """
    Return an initial sample collection, that will be incremented by the sampler.

    Args:
        key: PRNGKey
        num_live_points: the number of live points
        max_samples: the maximum number of samples
        model: the model to use.

    Returns:
        live_point_collection: the initial live point collection
        state: the initial state
    """

    def _repeat(a):
        return jnp.repeat(a[None], repeats=max_samples, axis=0, total_repeat_length=max_samples)

    sample_collection = StaticStandardSampleCollection(
        sender_node_idx=jnp.zeros(max_samples, dtype=mp_policy.index_dtype),
        log_L=jnp.full((max_samples,), jnp.inf, dtype=mp_policy.measure_dtype),
        U_samples=_repeat(model.U_placeholder),
        num_likelihood_evaluations=jnp.full((max_samples,), 0, dtype=mp_policy.count_dtype),
        phantom=jnp.full((max_samples,), False, dtype=jnp.bool_)
    )

    key, sample_key = random.split(key, 2)
    if mesh is not None:
        sharded_keys = tree_device_put(jax.random.split(sample_key, num_live_points), mesh, ('shard',))

        @partial(shard_map, mesh=mesh, in_specs=PartitionSpec('shard', ), out_specs=PartitionSpec('shard', ),
                 check_rep=False)
        def get_init_samples(local_keys):
            return draw_uniform_samples(
                keys=local_keys,
                model=model,
                method='scan'
            )

        live_point_samples = get_init_samples(sharded_keys)
    else:
        keys = jax.random.split(sample_key, num_live_points)
        live_point_samples = draw_uniform_samples(
            keys=keys,
            model=model,
            method='vmap'
        )
    live_point_collection = LivePointCollection(
        sender_node_idx=jnp.full((num_live_points,), 0, dtype=mp_policy.index_dtype),
        U_sample=live_point_samples.U_sample,
        log_L=live_point_samples.log_L,
        log_L_constraint=live_point_samples.log_L_constraint,
        num_likelihood_evaluations=live_point_samples.num_likelihood_evaluations
    )
    sort_indices = jnp.argsort(live_point_samples.log_L)
    live_point_collection = jax.tree_map(lambda x: x[sort_indices], live_point_collection)

    state = NestedSamplerState(
        key=key,
        next_sample_idx=jnp.asarray(0, mp_policy.index_dtype),
        num_samples=jnp.asarray(0, mp_policy.index_dtype),
        sample_collection=sample_collection
    )
    return live_point_collection, state


def create_init_termination_register() -> TerminationRegister:
    """
    Initialise the termination register.

    Returns:
        The initial termination register.
    """
    zero_count = jnp.asarray(0, mp_policy.count_dtype)
    init_evidence_calc = create_init_evidence_calc()
    return TerminationRegister(
        num_samples_used=zero_count,
        evidence_calc=init_evidence_calc,
        evidence_calc_with_remaining=init_evidence_calc,
        num_likelihood_evaluations=zero_count,
        log_L_contour=jnp.asarray(-jnp.inf, mp_policy.measure_dtype),
        efficiency=jnp.asarray(0., mp_policy.measure_dtype),
        plateau=jnp.asarray(False, jnp.bool_),
        no_seed_points=jnp.asarray(False, jnp.bool_),
        relative_spread=jnp.asarray(jnp.inf, mp_policy.measure_dtype),
        absolute_spread=jnp.asarray(jnp.inf, mp_policy.measure_dtype),
        peak_log_XL=jnp.asarray(-jnp.inf, mp_policy.measure_dtype)
    )



import jax.random
import jax.random
import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp

from jaxns import Prior, Model
from jaxns.experimental.global_optimisation import GlobalOptimisationTerminationCondition
from jaxns.experimental.public import GlobalOptimisation

tfpd = tfp.distributions

def xin_she_yang_1_problem_nd(ndims: int):

    def prior_model():
        z = yield Prior(tfpd.Uniform(low=-2. * jnp.pi * jnp.ones(ndims), high=2. * jnp.pi * jnp.ones(ndims)), name='z')
        return z

    def log_likelihood(z):
        return - jnp.sum(jnp.abs(z)) * jnp.exp(-jnp.sum(jnp.sin(z ** 2)))

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    optimum = 0. * jnp.ones(ndims)
    a_tol = 0.01
    log_L_tol = -0.01
    return model, optimum, a_tol, log_L_tol


def drop_wave_problem_nd(ndims: int):
    def prior_model():
        z = yield Prior(tfpd.Uniform(low=-5.12 * jnp.ones(ndims), high=5.12 * jnp.ones(ndims)), name='z')
        return z

    def log_likelihood(z):
        return (1 + jnp.cos(12. * jnp.sqrt(jnp.sum(z ** 2)))) / (0.5 * jnp.sum(z ** 2) + 2) - 1.

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    optimum = 0. * jnp.ones(ndims)
    a_tol = 0.01
    log_L_tol = -0.01
    return model, optimum, a_tol, log_L_tol


def run_performance_test_no_gradient_slice():
    (model, optimum, a_tol, log_L_tol) = drop_wave_problem_nd(5)
    keys = jax.random.split(jax.random.PRNGKey(42), 100)
    # gradient_slice=False, s=1, c=10, k=4, shell_frac=0.5:
    # mean_res=0.3747073956625793, success_prob=0.0
    # gradient_slice=False, s=1, c=50, k=4, shell_frac=0.5:
    # mean_res=0.3720548526096466, success_prob=0.019999999552965164
    # gradient_slice=False, s=1, c=100, k=4, shell_frac=0.5:
    # mean_res=0.35544749388887315, success_prob=0.08999999612569809
    # gradient_slice=False, s=5, c=10, k=24, shell_frac=0.5:
    # mean_res=0.37351151565613305, success_prob=0.019999999552965164
    # gradient_slice=False, s=5, c=50, k=24, shell_frac=0.5:
    # mean_res=0.33936949884649364, success_prob=0.11999999731779099
    # gradient_slice=False, s=5, c=100, k=24, shell_frac=0.5:
    # mean_res=0.31854922674033903, success_prob=0.17000000178813934
    # gradient_slice=False, s=10, c=10, k=49, shell_frac=0.5:
    # mean_res=0.35583731215755904, success_prob=0.07000000029802322
    # gradient_slice=False, s=10, c=50, k=49, shell_frac=0.5:
    # mean_res=0.32152508919560535, success_prob=0.1599999964237213
    # gradient_slice=False, s=10, c=100, k=49, shell_frac=0.5: <---- Best
    # mean_res=0.28122300728873983, success_prob=0.26999998092651367

    for s in [1, 5, 10]:
        for c in [10, 50, 100]:
            for k in [s * model.U_ndims - 1]:
                for shell_frac in [0.5]:
                    @jax.jit
                    def run(key):
                        go = GlobalOptimisation(
                            model,
                            gradient_slice=False,
                            shell_frac=shell_frac,
                            s=s,
                            k=k,
                            num_search_chains=model.U_ndims * c
                        )
                        results = go(
                            key=key,
                            term_cond=GlobalOptimisationTerminationCondition(log_likelihood_contour=log_L_tol),
                            finetune=False
                        )
                        return jnp.max(jnp.abs(results.solution[0] - optimum))

                    res = jax.block_until_ready(jnp.stack([run(key) for key in keys]))
                    mean_res = jnp.mean(res)
                    success = res < a_tol
                    success_prob = jnp.mean(success)
                    print(
                        f"gradient_slice={False}, s={s}, c={c}, k={k}, shell_frac={shell_frac}:\n"
                        f"mean_res={mean_res}, success_prob={success_prob}"
                    )


def run_performance_test_gradient_slice():
    (model, optimum, a_tol, log_L_tol) = xin_she_yang_1_problem_nd(2)
    # (model, optimum, a_tol, log_L_tol) = drop_wave_problem_nd(20)
    keys = jax.random.split(jax.random.PRNGKey(42), 100)

    # Can solve 100D dropwave with 100% success rate with about 275k likelihood evaluations, i.e. 2.75k per dimension
    # gradient_slice=True, s=1, c=1, k=0, shell_frac=0.5:
    # mean_res=1.7066995784509587e-05, success_prob=1.0, n_like=274194.42

    for s in [1, 2, 3, 4, 5]:
        for c in [1, 5, 10, 15]:
            for k in [s * model.U_ndims - 1]:
                for shell_frac in [0.5]:
                    @jax.jit
                    def run(key):
                        go = GlobalOptimisation(
                            model,
                            gradient_slice=True,
                            shell_frac=shell_frac,
                            s=s,
                            k=k,
                            num_search_chains=model.U_ndims * c
                        )
                        results = go(
                            key=key,
                            term_cond=GlobalOptimisationTerminationCondition(log_likelihood_contour=log_L_tol),
                            finetune=False
                        )
                        return jnp.max(jnp.abs(results.solution[0] - optimum)), results.num_likelihood_evaluations

                    res = []
                    n_like = []
                    for key in keys:
                        r, n = run(key)
                        res.append(r)
                        n_like.append(n)
                    res = jnp.stack(res)
                    n_like = jnp.stack(n_like)
                    mean_res = jnp.mean(res)
                    success = res < a_tol
                    success_prob = jnp.mean(success)
                    print(
                        f"gradient_slice={True}, s={s}, c={c}, k={k}, shell_frac={shell_frac}:\n"
                        f"mean_res={mean_res}, success_prob={success_prob}, n_like={jnp.mean(n_like)}"
                    )


if __name__ == '__main__':
    # run_performance_test_no_gradient_slice()
    run_performance_test_gradient_slice()



import dataclasses
import warnings
from typing import NamedTuple, Tuple, Any

import jax
from jax import random, numpy as jnp, lax

from jaxns.framework.bases import BaseAbstractModel
from jaxns.internals.mixed_precision import mp_policy
from jaxns.internals.types import IntArray, UType
from jaxns.internals.types import PRNGKey, FloatArray
from jaxns.nested_samplers.common.types import Sample
from jaxns.samplers.abc import EphemeralState
from jaxns.samplers.bases import BaseAbstractRejectionSampler
from jaxns.samplers.multi_ellipsoid.multi_ellipsoid_utils import ellipsoid_clustering, MultEllipsoidState
from jaxns.samplers.multi_ellipsoid.multi_ellipsoid_utils import sample_multi_ellipsoid

__all__ = [
    'MultiEllipsoidalSampler'
]


@dataclasses.dataclass(eq=False)
class MultiEllipsoidalSampler(BaseAbstractRejectionSampler[MultEllipsoidState]):
    """
    Uses a multi-ellipsoidal decomposition of the live points to create a bound around regions to sample from.

    Inefficient for high dimensional problems, but can be very efficient for low dimensional problems.
    """
    model: BaseAbstractModel
    depth: int
    expansion_factor: float

    def __post_init__(self):
        warnings.warn(
            "MultiEllipsoidalSampler does not give consistent results/performance. Consider `UniDimSliceSampler`.")
        if self.depth < 0:
            raise ValueError(f"depth {self.depth} must be >= 0")
        if self.expansion_factor <= 0.:
            raise ValueError(f"expansion_factor {self.expansion_factor} must be > 0")

    def num_phantom(self) -> int:
        return 0

    def _pre_process(self, ephemeral_state: EphemeralState) -> Any:
        return ellipsoid_clustering(
            key=ephemeral_state.key,
            points=ephemeral_state.live_points_collection.U_sample,
            log_VS=ephemeral_state.termination_register.evidence_calc_with_remaining.log_X_mean,
            max_num_ellipsoids=self.max_num_ellipsoids,
            method='em_gmm'
        )

    def _post_process(self, ephemeral_state: EphemeralState,
                      sampler_state: Any) -> Any:
        return sampler_state

    @property
    def max_num_ellipsoids(self):
        return 2 ** self.depth

    def _get_sample(self, key: PRNGKey, log_L_constraint: FloatArray, sampler_state: MultEllipsoidState) -> Tuple[
        Sample, Sample]:
        def _sample_multi_ellipsoid(key: PRNGKey) -> UType:
            _, U = sample_multi_ellipsoid(
                key=key,
                mu=sampler_state.params.mu,
                radii=sampler_state.params.radii * jnp.asarray(self.expansion_factor, mp_policy.measure_dtype),
                rotation=sampler_state.params.rotation,
                unit_cube_constraint=True
            )
            return U

        class CarryState(NamedTuple):
            key: PRNGKey
            U: FloatArray
            log_L: FloatArray
            num_likelihood_evals: IntArray

        def cond(carry: CarryState):
            return carry.log_L <= log_L_constraint

        def body(carry: CarryState):
            key, sample_key = random.split(carry.key, 2)
            point_U = _sample_multi_ellipsoid(key=sample_key)
            log_L = self.model.forward(U=point_U)
            num_likelihood_evals = carry.num_likelihood_evals + jnp.ones_like(carry.num_likelihood_evals)
            # backoff by one e-fold per attempt after efficiency threshold reached
            return CarryState(
                key=key,
                U=point_U,
                log_L=log_L,
                num_likelihood_evals=num_likelihood_evals
            )

        key, sample_key = random.split(key, 2)
        point_U = _sample_multi_ellipsoid(key=sample_key)
        init_carry_state = CarryState(
            key=key,
            U=point_U,
            log_L=self.model.forward(point_U),
            num_likelihood_evals=jnp.asarray(1, mp_policy.count_dtype)
        )

        final_carry = lax.while_loop(
            cond_fun=cond,
            body_fun=body,
            init_val=init_carry_state
        )

        sample = Sample(
            U_sample=final_carry.U,
            log_L_constraint=log_L_constraint,
            log_L=final_carry.log_L,
            num_likelihood_evaluations=final_carry.num_likelihood_evals
        )
        # TODO: could use rejected samples, not as phantom because they don't satisfy constraint, but for ML apps
        phantom_samples = jax.tree.map(lambda x: jnp.zeros((0,) + x.shape, x.dtype), sample)  # [k, D]
        return sample, phantom_samples



import base64
import importlib

import jax
import jax.numpy as jnp
import numpy as np


def isinstance_namedtuple(obj) -> bool:
    """
    Check if object is a namedtuple.

    Args:
        obj: object

    Returns:
        bool
    """
    return (
            isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields')
    )


def issubclass_namedtuple(cls):
    """
    Check if the type object is a subclass of a namedtuple.
    """
    base_types = cls.__mro__  # Get the method resolution order of the class
    return any(hasattr(base, '_fields') and hasattr(base, '_asdict') for base in base_types)


def serialise_namedtuple(obj):
    if isinstance_namedtuple(obj):
        class_name = f"{obj.__class__.__module__}.{obj.__class__.__name__}"
        return {'type': '__namedtuple__', '__class__': class_name,
                '__data__': {k: serialise_namedtuple(v) for k, v in obj._asdict().items()}}
    elif isinstance(obj, np.ndarray):
        return serialise_ndarray(obj)
    elif isinstance(obj, jax.Array):
        return serialise_jax_ndarray(obj)
    elif isinstance(obj, (list, tuple)):
        return [serialise_namedtuple(v) for v in obj]
    elif isinstance(obj, dict):
        return {k: serialise_namedtuple(v) for k, v in obj.items()}
    else:
        return obj


def deserialise_namedtuple(obj):
    if isinstance(obj, dict) and 'type' in obj and obj['type'] == '__namedtuple__':
        class_path = obj['__class__']
        module_name, class_name = class_path.rsplit('.', 1)
        module = importlib.import_module(module_name)
        class_ = getattr(module, class_name)
        return class_(**{k: deserialise_namedtuple(v) for k, v in obj['__data__'].items()})
    elif isinstance(obj, dict) and 'type' in obj and obj['type'] == '__ndarray__':
        return deserialise_ndarray(obj)
    elif isinstance(obj, dict) and 'type' in obj and obj['type'] == '__jax_ndarray__':
        return deserialise_jax_ndarray(obj)
    elif isinstance(obj, list):
        return [deserialise_namedtuple(v) for v in obj]
    elif isinstance(obj, tuple):
        return tuple(deserialise_namedtuple(v) for v in obj)
    elif isinstance(obj, dict):
        return {k: deserialise_namedtuple(v) for k, v in obj.items()}
    return obj


def serialise_ndarray(obj):
    if isinstance(obj, np.ndarray):
        data_bytes = obj.tobytes()
        bytes_base64 = base64.b64encode(data_bytes).decode('utf-8')
        return {'type': '__ndarray__', '__dtype__': str(obj.dtype), '__data__': bytes_base64, '__shape__': obj.shape}
    return obj


def deserialise_ndarray(obj):
    if isinstance(obj, dict) and obj.get('type') == '__ndarray__':
        bytes_base64 = obj['__data__']
        data_bytes = base64.b64decode(bytes_base64)
        # make array from bytes and give correct dtype and shape
        return np.frombuffer(data_bytes, dtype=obj['__dtype__']).reshape(obj['__shape__'])
    return obj


def serialise_jax_ndarray(obj):
    if isinstance(obj, jax.Array):
        data_bytes = np.asarray(obj).tobytes()
        bytes_base64 = base64.b64encode(data_bytes).decode('utf-8')
        return {'type': '__jax_ndarray__', '__dtype__': str(obj.dtype), '__data__': bytes_base64,
                '__shape__': obj.shape}
    return obj


def deserialise_jax_ndarray(obj):
    if isinstance(obj, dict) and obj.get('type') == '__jax_ndarray__':
        bytes_base64 = obj['__data__']
        data_bytes = base64.b64decode(bytes_base64)
        # make array from bytes and give correct dtype and shape
        return jnp.frombuffer(data_bytes, dtype=obj['__dtype__']).reshape(obj['__shape__'])
    return obj
